#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#
from typing import List, Union, Callable, Optional

import serial  # type: ignore
import time
import threading
from htf.powersupply.core import (
    ChannelVoltage,
    ActualChannelVoltage,
    ChannelCurrent,
    ActualChannelCurrent,
)
import logging

# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class _HMPSeriesAbstraction:
    """
    An SCPI interface for all Rohde & Schwarz HMP power supplies.
    """

    NUMBER_OF_CHANNELS: Optional[int] = None
    IDNS: List[bytes] = []

    voltage1 = ChannelVoltage(1)
    voltage2 = ChannelVoltage(2)
    voltage3 = ChannelVoltage(3)
    voltage4 = ChannelVoltage(4)

    actual_voltage1 = ActualChannelVoltage(1)
    actual_voltage2 = ActualChannelVoltage(2)
    actual_voltage3 = ActualChannelVoltage(3)
    actual_voltage4 = ActualChannelVoltage(4)

    current1 = ChannelCurrent(1)
    current2 = ChannelCurrent(2)
    current3 = ChannelCurrent(3)
    current4 = ChannelCurrent(4)

    actual_current1 = ActualChannelCurrent(1)
    actual_current2 = ActualChannelCurrent(2)
    actual_current3 = ActualChannelCurrent(3)
    actual_current4 = ActualChannelCurrent(4)

    def __init__(
        self,
        port: str,
        baudrate: int = 9600,
        parity: str = serial.PARITY_NONE,
        stopbits: Union[int, float] = serial.STOPBITS_ONE,
        timeout: Union[int, float] = 1.0,
        command_delay: Union[int, float] = 0.05,
        wait_after_set_voltage: bool = False,  # fast changes in setting voltage is possible, but be aware, that the
        # device needs time for reaching setpoint level
        wait_time: Union[int, float] = 0.5,
    ) -> None:
        self._command_delay = command_delay
        self._wait_after_set_voltage = wait_after_set_voltage
        self._wait_time = wait_time
        self._serial = serial.serial_for_url(
            port,
            baudrate=baudrate,
            parity=parity,
            stopbits=stopbits,
            timeout=timeout,
            write_timeout=timeout,
            rtscts=True,
        )
        self._lock = threading.Lock()

        self._check_idn()

    def close(self) -> None:
        """
        Close serial connection.
        """
        if hasattr(self, "_serial"):
            self._serial.close()

    def __del__(self) -> None:
        self.close()

    def _check_idn(self) -> None:
        device_idn = self.get_idn()

        # loop over all possible IDNs and raise exception if no IDN matches

        for idn in self.IDNS:
            if device_idn.startswith(idn):
                return

        idns = '"' + "', '".join(map(lambda x: x.decode(errors="ignore"), self.IDNS)) + '"'

        raise AssertionError(
            f"Expected IDN to start with {idns} but found "
            f"'{device_idn.decode(errors='ignore')}'. "
            f"Select the right powersupply abstractor." % (b", ".join(self.IDNS), device_idn)
        )

    def _set_voltage_for_channel_generator(self, channel: int) -> Callable[[Union[int, float]], None]:
        def _set_voltage_for_channel(voltage: Union[int, float]) -> None:
            self.select_channel(channel)
            self.set_voltage(voltage, auto_enable=True)

        return _set_voltage_for_channel

    def _get_voltage_for_channel_generator(self, channel: int) -> Callable[[], None]:
        def _get_voltage_for_channel() -> Union[int, float]:
            self.select_channel(channel)
            return self.get_voltage()  # type: ignore

        return _get_voltage_for_channel  # type: ignore

    def _get_actual_voltage_for_channel_generator(self, channel: int) -> Callable[[], None]:
        def _get_actual_voltage_for_channel() -> Union[int, float]:
            self.select_channel(channel)
            return self.get_actual_voltage()  # type: ignore

        return _get_actual_voltage_for_channel  # type: ignore

    def _set_current_for_channel_generator(self, channel: int) -> Callable[[Union[int, float]], None]:
        def _set_current_for_channel(current: Union[int, float]) -> None:
            self.select_channel(channel)
            self.set_current(current)

        return _set_current_for_channel

    def _get_current_for_channel_generator(self, channel: int) -> Callable[[], None]:
        def _get_current_for_channel() -> Union[int, float]:
            self.select_channel(channel)
            return self.get_current()  # type: ignore

        return _get_current_for_channel  # type: ignore

    def _get_actual_current_for_channel_generator(self, channel: int) -> Callable[[], None]:
        def _get_actual_current_for_channel() -> Union[int, float]:
            self.select_channel(channel)
            return self.get_actual_current()  # type: ignore

        return _get_actual_current_for_channel  # type: ignore

    def _query(self, commands: List[bytes], response_length: int = 0) -> bytes:
        with self._lock:
            self._serial.flushInput()

            for command in commands:
                command = b"\r\n" + command + b"\r\n"
                logger.debug(f"send: '{command.decode(errors='ignore')}'")
                self._serial.write(command)
                time.sleep(self._command_delay)

            response = b""
            received = 0

            while received < response_length:
                b = self._serial.read(1)
                logger.debug("received '%s'" % b)
                if b in [b"", b"\n"]:
                    break
                response += b
                received += 1

        return response

    def get_idn(self) -> bytes:
        """
        Get the identifier.

        Returns:
            str: the identifier of the device
        """
        idn = self._query([b"*IDN?"], response_length=100)
        return idn

    def select_channel(self, channel: int) -> None:
        """
        Select a channel.

        Args:
            channel: the channel to be selected (``1 .. 4``)
        """
        assert 1 <= channel <= 4
        ret = self._query([b"INST OUT%i" % channel, b"INST?"], response_length=5)
        assert ret == b"OUTP%i" % channel

    def select_output_channel(self) -> None:
        """
        Select currently selected channel for general output.
        """
        self._query([b"OUTP:SEL ON"], response_length=0)

    def deselect_output_channel(self) -> None:
        """
        Deselect currently selected channel for general output.
        """
        self._query([b"OUTP:SEL OFF"], response_length=0)

    def enable_output(self) -> None:
        """
        Enable the output of the currently selected channel.
        """
        ret = self._query([b"OUTP ON", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def disable_output(self) -> None:
        """
        Disable the output of the currently selected channel.
        """
        ret = self._query([b"OUTP OFF", b"OUTP?"], response_length=1)
        assert ret == b"0"

    def enable_general_output(self) -> None:
        """
        Enable the general output.
        """
        ret = self._query([b"OUTP:GEN ON", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def disable_general_output(self) -> None:
        """
        Enable the general output.
        """
        ret = self._query([b"OUTP:GEN OFF", b"OUTP?"], response_length=1)
        assert ret == b"1"

    def store_state(self, location: int) -> None:
        """
        Save the current instrument state to the specified storage location.

        Args:
            location: the storage location to store the instrument state to.
        """
        assert 0 <= location < 10, "location must be between 0 and 9"
        self._query([b"*SAV %i" % location], response_length=0)

    def recall_state(self, location: int) -> None:
        """
        Recall the instrument state from the specified storage location.

        Args:
            location: the storage location to load the instrument state from.
        """
        assert 0 <= location < 10, "location must be between 0 and 9"
        self._query([b"*RCL %i" % location], response_length=0)

    def set_voltage(self, voltage: Union[int, float], auto_enable: bool = False) -> None:
        """
        Set the voltage on the currently selected channel.

        Args:
            voltage: the voltage to be set.
            auto_enable=False: if set to ``True`` the output is enabled automatically.
        """
        if voltage is not None:
            ret = self._query([b"VOLT %.3f" % voltage, b"VOLT?"], response_length=6)
            expected_voltage = b"%.3f" % voltage
            assert ret == expected_voltage
            if auto_enable is True:
                self.enable_output()
            if self._wait_after_set_voltage is True:
                logger.debug("wait after set voltage to %f V for %f s" % (voltage, self._wait_time))
                time.sleep(self._wait_time)
        else:
            if auto_enable is True:
                self.disable_output()

    def get_voltage(self) -> Union[int, float]:
        """
        Get the voltage of the currently selected channel.

        Returns:
            float: the voltage of the currently selected channel
        """
        ret = self._query([b"VOLT?"], response_length=6)
        return float(ret)

    def get_actual_voltage(self) -> float:
        """
        Get the measured voltage of the currently selected channel.

        Returns:
            float: the measured voltage of the currently selected channel
        """
        ret = self._query([b"MEAS:VOLT?"], response_length=6)
        return float(ret)

    def set_current(self, current: Union[int, float], read_back: bool = True) -> None:
        """
        Set the current of the currently selected channel

        Args:
            current: the current to be set
            read_back=True: if True the set current is read back and compared to the expected current
        """
        current = round(current, 3)
        logger.debug("_set_current: %f", current)
        ret = self._query([b"CURR %.3f" % current, b"CURR?"], response_length=6)
        if read_back:
            if current < 10:
                expected_current = b"%.4f" % current  # 5 significant places
            else:
                expected_current = b"%.3f" % current  # 5 significant places
            logger.debug(f"expected_current {expected_current.decode(errors='ignore')}")
            assert ret == expected_current, (
                f"read back current mismatch: "
                f"expected '{expected_current.decode(errors='ignore')}' "
                f"but '{ret.decode(errors='ignore')}' was recieved"
            )

    def get_current(self) -> float:
        """
        Get the current of the currently selected channel.

        Returns:
            float: the current of the currently selected channel
        """
        ret = self._query([b"CURR?"], response_length=6)
        return float(ret)

    def get_actual_current(self) -> float:
        """
        Get the measured current of the currently selected channel.

        Returns:
            float: the measured current of the currently selected channel
        """
        ret = self._query([b"MEAS:CURR?"], response_length=6)
        return float(ret)


class HMP2020(_HMPSeriesAbstraction):
    """
    HMP2020: 1 × 0 V to 32 V/0 A to 10 A; 1 × 0 V to 32 V/0 A to 5 A (188 W)
    """

    NUMBER_OF_CHANNELS = 2
    IDNS = [b"HAMEG,HMP2020,", b"ROHDE&SCHWARZ,HMP2020,"]


class HMP2030(_HMPSeriesAbstraction):
    """
    HMP2030: 3 × 0 V to 32 V/0 A to 5 A (188 W)
    """

    NUMBER_OF_CHANNELS = 3
    IDNS = [b"HAMEG,HMP2030,", b"ROHDE&SCHWARZ,HMP2030,"]


class HMP4030(_HMPSeriesAbstraction):
    """
    HMP4030: 3 × 0 V to 32 V/0 A to 10 A (384 W)
    """

    NUMBER_OF_CHANNELS = 3
    IDNS = [b"HAMEG,HMP4030,", b"ROHDE&SCHWARZ,HMP4030,"]


class HMP4040(_HMPSeriesAbstraction):
    """
    HMP4040: 4 × 0 V to 32 V/0 A to 10 A (384 W)
    """

    NUMBER_OF_CHANNELS = 4
    IDNS = [b"HAMEG,HMP4040,", b"ROHDE&SCHWARZ,HMP4040,"]
